#ifndef rest_MATRIX_H
#define rest_MATRIX_H
// ================ Include-Files ==== ==============================
#include "pch.h"
// #include "Matrizen.h"
// =============== Klassen-Deklaration: MATRIX  =========================================
// MATRIX-Klasse fuer Matrizenoperationen vom Typ double
// Rechteckmatrix mat(i,j,k,l) wird spaltenweise gespeichert !!!
// Indizierung von i,j,k,l beginnt bei 0 !!!
// vgl. U. Breymann: C++. Carl Hanser Verlag (1993), S.186 ff
// ======================================================================================
// Copyright:: Univ.-Prof. Dr.-Ing. Udo F. Meissner, Erkrath / Darmstadt
// experimentelles Lehr- und Studienprogramm - kommerzielle Nutzung nicht gestattet
// Der Autor bernimmt keine Haftung fr Unvollstndigkeiten und Maengel der Algorithmen!
// Version:  1995
// Compiler: C++
// ========= Deklaration der Prototypen (header file) ===============
    class MATRIX : public MATRIZEN
{
    // ==============================================================
    //                 Oeffentliche Mitgliedsfunktionen
    //                 ================================
    //
    // **************************************************************
    // *   **** dieser Methodenteil ist im Buch dokumentiert ****   * 
    // **************************************************************
    //
    // ==============================================================
    // Oeffentliche, virtuelle Mitgliedsfunktionen (Gleichungssysteme)
    // ===============================================================
       public:
    // =======
    //
    // Symmetrische Koeffizientenmatrix nach Gauss staffeln
    // -> Amat.nach_gauss_staffeln(Umat)
    // =====================================================
    // Symmetrische Koeffizientenmatrix
    // nach Gauss in die gestaffelte Dreiecksform ueberfuehren
    // A(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // U(mk,mk) = Matrix fuer die gestaffelte Form der Obermatrix U
    //            (U darf gleich A sein) 
    // mk       = Zeilen/Spalten von A, U
    // Ergebnis: Determinante der Matrix A
    // -----------------------------------------------------------------
    double nach_gauss_staffeln(MATRIZEN& U);
    //
    // Symmetrische Koeffizientenmatrix nach Cholesky staffeln
    // -> Amat.nach_cholesky_staffeln(Umat)
    // =======================================================
    // Symmetrische und positiv definite Koeffizientenmatrix
    // nach Cholesky in die gestaffelte Dreiecksform ueberfuehren
    // A(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // U(mk,mk) = Matrix fuer die gestaffelte Form der Obermatrix U
    //            (U darf gleich A sein) 
    // mk       = Zeilen/Spalten von A, U
    // Ergebnis: Determinante der Matrix A
    // -------------------------------------------------------------
    double nach_cholesky_staffeln(MATRIZEN& U);
    //
    // Unsymmetrische Koeffizientenmatrix nach Gauss staffeln
    // -> Amat.nach_gauss_staffeln(Umat,Lmat)
    // ======================================================
    // Unsymmetrische Koeffizientenmatrix
    // nach Gauss in die gestaffelte Dreiecksform ueberfuehren
    // A(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // U(mk,mk) = Matrix fuer die gestaffelte Form der Obermatrix U
    //            (U darf gleich A sein) 
    // L(mk,mk) = Matrix fuer die gestaffelte Form der Untermatrix L
    //            (L darf gleich U sein)
    // mk       = Zeilen/Spalten von K, U, L
    // Ergebnis: Determinante der Matrix A
    // -----------------------------------------------------------------
    double nach_gauss_staffeln(MATRIZEN& U, MATRIZEN& L);
    //
    // Rechte Seite nach Cholesky oder Gauss staffeln 
    // -> Rmat.staffeln(Lmat,Ymat,symdef)
    // ==================================================================
    // Rechte Seite eines Gleichungssystems nach
    // Cholesky oder Gauss in die gestaffelte Rechteckform ueberfuehren
    // R(mb,zr) = Matrix mit voll abgespeicherten Rechten Seiten
    // L(mb,mb) = Matrix mit gestaffelter Koeffizientenmatrix
    //            (Untermatrix L bei Gauss unsymmetrisch)
    // Y(mb,zr) = Matrix fuer gestaffelte Form der Rechten Seiten
    //            (Y darf gleich R sein)
    // mb       = Zeilenzahl von L,R,Y
    // zr       = Spaltenzahl der Rechten Seiten R,Y
    // symdef= 1 : nach Cholesky staffeln
    //       = 0 : nach Gauss symmetrisch   staffeln 
    //       =-1 : nach Gauss unsymmetrisch staffeln
    // -------------------------------------------------------------------
    void staffeln(MATRIZEN& L, MATRIZEN& Y, int symdef = -1);
    //
    // Loesungen nach Cholesky oder Gauss ermitteln
    // -> Xmat.loesungen(Umat,Ymat,symdef)
    // =============================================
    // Loesungen fuer Gleichungsystem nach Cholesky (symmetrisch) oder
    // nach Gauss (symmetrisch oder unsymmetrisch) aus der gestaffelten
    // Koeffizientenmatrix und den gestaffelten Rechten Seiten ermitteln
    // X(mb,zr) = Matrix fuer voll abgespeicherte Loesungen
    // U(mb,mb) = Matrix mit gestaffelter Koeffizientenmatrix
    //            (Obermatrix U bei Gauss unsymmetrisch)
    // Y(mb,zr) = Matrix mit gestaffelter Form der Rechten Seiten
    //            (Y darf gleich X sein)
    // mb       = Zeilenzahl von U,X,Y
    // zr       = Spaltenzahl der Loesungen Seiten X,Y
    // symdef= 1 : nach Cholesky berechnen
    //       = 0 : nach Gauss symmetrisch   berechnen
    //       =-1 : nach Gauss unsymmetrisch berechnen
    // -----------------------------------------------------------------
    //*** void loesungen(MATRIZEN& U, MATRIZEN& Y, int symdef = -1);    // Korrektur 01/2023
    MATRIZEN& loesungen(MATRIZEN& U, MATRIZEN& Y, int symdef = -1);
    //
    // Inverse einer   symmetrischen Matrix nach Cholesky oder Gauss
    // Inverse einer unsymmetrischen Matrix nach Gauss bilden
    // -> Amat.invertieren(Zmat,symdef)
    // =============================================================
    // Koeffizientenmatrix
    // nach Cholesky (nur positiv definite Matrix) oder Gauss invertieren
    // A(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // Z(mk,mk) = Matrix fuer die Inverse
    //            (Z darf gleich A sein) 
    // mk       = Zeilen/Spalten von A, Z
    // symdef   =  1 Cholesky
    //          =  0 Gauss symmetrisch
    //          = -1 Gauss unsymmetrisch
    // Ergebnis: Determinante der Matrix A
    // --------------------------------------------------------------------
    double invertieren(MATRIZEN& Z, int symdef = -1);
    // 
    // Lagerbedingungen in Koeffizientenmatrix einarbeiten
    // -> Kmat.lagern(Smat,Amat)
    // ===================================================
    // Lagerbedingungen in Koeffizientenmatrix einarbeiten 
    // K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // S(mk)    = Vektor mit Lagerangaben
    //            -1 = Starre Lagerung
    //             0 = keine  Lagerung
    //            >0 = Federkonstante auf der Hauptdiagonale addieren
    // A(mk,mk) = Matrix fuer die gelagerte Koeffizientenmatrix
    //            (A darf gleich K sein)
    // mk       = Zeilen/Spalten von K, A
    //            Zeilen von S
    //---------------------------------------------------------
    void lagern(MATRIZEN& S, MATRIZEN& A);
    //     
    // ==============================================================
    //                 Geschuetzte Mitgliedsfunktionen
    //                 ===============================
       protected:
    // ==========
    //
    // Matrixelement -> mat.matrix_element(i,j,k,l) vergleiche mat(i,j,k,l)
    // ====================================================================
    double& matrix_element(int i = 0, int j = 0, int k = 0, int l = 0);
    // ====================================================================
    //
    // Referenztyp wandeln -> MATRIX &mat=referenz(MATRIZEN &mat_)
    // ===========================================================
    MATRIX& referenz(MATRIZEN& mat);
    //
    //                 Geschuetzte, virtuelle Mitgliedsfunktionen
    //                 ==========================================
       protected:
    // ==========
    //
    // Adresse von Matrixelement -> mat.adresse(i,j,k,l)
    //==================================================
    double* adresse(int i = 0, int j = 0, int k = 0, int l = 0);
    //
    // Alle Matrix-Elemente belegen -> mat.zuweisen(zahl,zweig)
    //                                 (nur Matrizeninhalt kopieren)
    // =============================================================
    // Zweig= 0 -> mat=    zahl
    //      = 2 -> mat=mat+zahl
    //      = 4 -> mat=mat-zahl
    // ------------------------
    void zuweisen(double zahl, int zweig = 0);
    //
    // Matrix zuweisen -> mat1.zuweisen(mat2,zweig)
    //                    (nur Matrizeninhalt kopieren)
    // ================================================
    // Zweig= 0 -> mat1=    mat2
    //      = 2 -> mat1=mat+mat2
    //      = 4 -> mat1=mat-mat2
    void zuweisen(MATRIZEN& mat2, int zweig = 0);
    //
    // Temporaere Matrix loeschen -> mat.temporaere_matrix_loeschen()
    // ==============================================================
    void temporaere_matrix_loeschen();
    //
    // =====================================================================
    //                 Private, virtuelle Mitgliedsfunktionen
    //                 ======================================
       private:
    // ========
    void erzeugen(int ni = 1, int nj = 1, int nk = 1, int nl = 1,
                                        int neutyp = 1,double* C_array=NULL);
    //
    // Matrix kopieren -> kopieren(&mat,neutyp)
    // ========================================
    // bei neutyp=-1 Matrix-Deskriptor, neuen Speicher anfordern (nur privat !!!)
    // bei neutyp= 0 Matrix-Deskriptor, alten Speicher benutzen  (nur privat !!!)
    // bei neutyp= 1 Matrix-Deskriptor, neuen Speicher anfordern, Matrix kopieren
    // C_array vorinitialisiert mittels Initialisierungsliste C_feld [] = { ... }
    // ---------------------------------------------------------------------------
    void kopieren(MATRIZEN &mat,int neutyp=1);
    //
    // temporaere Matrix erzeugen -> temporaere_matrix_erzeugen(ni,nj,nk,nl,neutyp)
    // ============================================================================
    // bei neutyp=-1 temporaere Matrix erzeugen
    // bei neutyp= 1 permanente Matrix erzeugen
    // ni, nj, nk, nl: Dimensionen der 4D-Matrix
    // -----------------------------------------
    MATRIZEN *temporaere_matrix_erzeugen(int ni,int nj=1,int nk=1,int nl=1,
                                                                  int neutyp=-1);
    //
    // temporaere Matrix als Kopie erzeugen
    //                                  -> temporaere_matrix_erzeugen(&mat,neutyp)
    // ===========================================================================
    // Kopie der Matrix mat:
    // bei neutyp=-1 Matrix-Deskriptor, neuen Speicher anfordern (nur privat !!!)
    // bei neutyp= 0 Matrix-Deskriptor, alten Speicher benutzen  (nur privat !!!)
    // bei neutyp= 1 Matrix-Deskriptor, neuen Speicher anfordern, Matrix kopieren
    // ---------------------------------------------------------------------------
    MATRIZEN *temporaere_matrix_erzeugen(MATRIZEN &mat,int neutyp=-1);
    //
    // Matrix-Dimensionen temporaer aendern -> mat.dimension_temporaer(i,j,k,l)
    // ========================================================================
    // Das Zusammenfassen von Indizes (Umdimensionierung) wird eigentlich nur
    //                    fuer die Matrizenmultiplikation mat1*mat2 benoetigt.
    // Daher gilt die Umdimensionierung nur bis zur naechsten Matrizenoperartion !
    // Danach werden die Originaldimensionen wiederhergestellt !!!
    // ---------------------------------------------------------------------------
    void dimension_temporaer(int ni=1,int nj=1,int nk=1,int nl=1);
    //
    // Teilmatrix heraus -> ...=mat1.teil_heraus(mat2, i,j,k,l)
    // ========================================================
    // teil_heraus holt ab Index mat2(i,j,k,l) aufeinanderfolgende Matrixelemente
    // heraus, die in der Teilmatrix mat1(ni,nj,nk,nl) gespeichert werden
    // ---------------------------------------------------------------------------
    void teil_heraus(MATRIZEN &mat2, int i=0,int j=0,int k=0,int l=0);
    //
    // Teilmatrix hinein -> mat.teil_hinein(i,j,k,l)=...
    // =================================================
    // teil_hinein erlaubt das Einspeichern der Teilmatrix ab Index mat(i,j,k,l)
    // ---------------------------------------------------------------------------
    void teil_hinein(int i=0, int j=0, int k=0, int l=0);
    //
    // Untermatrix heraus -> ...=mat1.unter_heraus(mat2, i,j,k,l)
    // ==========================================================
    // unter_heraus holt ab Index mat2(i,j,k,l)
    //                                    die Untermatrix mat1(ni,nj,nk,nl) heraus
    // ---------------------------------------------------------------------------
    void unter_heraus(MATRIZEN &mat2, int i=0,int j=0,int k=0,int l=0);
    //
    // Untermatrix hinein -> mat.unter_hinein(i,j,k,l)=...
    // ===================================================
    // unter_hinein erlaubt das Einspeichern der Untermatrix ab Index mat(i,j,k,l)
    // ---------------------------------------------------------------------------
    void unter_hinein(int i=0,int j=0,int k=0,int l=0);
    //
    // Matrix-Ausgabe -> mat.ausgeben(s)
    // =================================
    void ausgeben(ostream &s);
    //
    // Matrix-Eingabe -> mat.einlesen(s)
    // =================================
    void einlesen(istream &s);
    //
    // Addition -> mat1+mat2 bzw. mat1+vek2 bzw. vek1+mat2
    // ===================================================
    void addieren(MATRIZEN &mat1,MATRIZEN &mat2);
    //
    // Subtraktion -> mat1-mat2 bzw. mat1-vek2 bzw. vek1-mat2
    // ======================================================
    void subtrahieren(MATRIZEN &mat1,MATRIZEN &mat2);
    //
    // Matrizenmultiplikation -> matc(i,j,k,l) = mat1(i,s,j) * mat2(k,s,l)
    // zweig = 0 -> matc = mat1 * mat2 / =  1 -> matc = matc + mat1 * mat2 /
    //                                   = -1 -> matc = matc - mat1 * mat2 
    // ===========================================================================
    void multiplizieren(MATRIZEN &mat1,MATRIZEN &mat2, int zweig=0);
    //
    // ===========================================================================
    //                 Private Mitgliedsfunktionen
    //                 ===========================
       private:
    // ========
    // Konstruktor -> MATRIX mat(ni,nj,nk,nl,neutyp)
    // =====================================================
    // ni, nj, nk, nl: Dimensionen der 4D-Matrix
    // C_array vorinitialisiert mittels Initialisierungsliste C_feld [] = { ... }  
    // bei neutyp=-1 temporaere Matrix erzeugen
    // bei neutyp= 1 permanente Matrix erzeugen
    // --------------------------------------------------------------------------
    MATRIX(int ni,int nj,int nk,int nl,int neutyp);
    //
    // Kopier-Konstruktor -> MATRIX mat1(&mat2,neutyp)
    // ===============================================
    // bei neutyp=-1 Matrix-Deskriptor, neuen Speicher anfordern (nur privat !!!)
    // bei neutyp= 0 Matrix-Deskriptor, alten Speicher benutzen  (nur privat !!!)
    // bei neutyp= 1 Matrix-Deskriptor, neuen Speicher anfordern, Matrix kopieren
    // ---------------------------------------------------------------------------
    MATRIX(MATRIX &mat2,int neutyp);
    //
    // Index ijkl von Matrixelement -> mat.index(i,j,k,l)
    // ==================================================
    int index(int i=0, int j=0, int k=0, int l=0);
    //
    // 2D-Indizes von Matrizenelement -> mat.index_2d(ij, dimi,dimj, i,j)
    // Eingabe: 1D-Index ij, Dimensionen dimi,dimj, Ausgabe: 2D-Indizes i,j
    // ====================================================================
    void index_2d(int ij,int dimi,int dimj,int &i,int &j);
    //
    // 3D-Indizes von Matrizenelement -> mat.index_3d(ijk, dimi,dimj,dimk, i,j,k)
    // Eingabe: 1D-Index ijk,Dimensionen dimi,dimj,dimk, Ausgabe: 3D-Indizes i,j,k
    // ===========================================================================
    void index_3d(int ijk,int dimi,int dimj,int dimk,int &i,int &j,int &k);
    //
    // 4D-Indizes von Matrizenelement -> mat.index_4d(ijkl, i,j,k,l)
    // Eingabe: 1D-Index ijkl, Ausgabe: 4D-Indizes i,j,k,l
    // =============================================================
    void index_4d(int ijkl,int &i,int &j,int &k,int &l);
    //
    // ========================== Gleichungssysteme ==============================
    //    
    // Symmetrische Koeffizientenmatrix nach Gauss staffeln -> S3151(Kmat,Cmat,mk)
    // ===========================================================================
    // Symmetrische Koeffizientenmatrix
    // nach Gauss in die gestaffelte Dreiecksform (Untermatrix C) ueberfuehren
    // K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // C(mk,mk) = Matrix fuer die gestaffelte Form
    //            (C darf gleich K sein) 
    // mk       = Zeilen/Spalten von K, C
    // Ergebnis: Determinante der Matrix K
    // -------------------------------------------------------------
    double S3151(MATRIZEN& K, MATRIZEN& C, int mk);
    //
    // Symmetrische Koeffizientenmatrix nach Cholesky staffeln
    //                                                      -> S3152(Kmat,Bmat,mk)
    // ===========================================================================
    // Symmetrische und positiv definite Koeffizientenmatrix
    // nach Cholesky in die gestaffelte Dreiecksform ueberfuehren
    // K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // B(mk,mk) = Matrix fuer die gestaffelte Form
    //            (B darf gleich K sein) 
    // mk       = Zeilen/Spalten von K, B
    // Ergebnis: Determinante der Matrix K
    // -------------------------------------------------------------
    double S3152(MATRIZEN& K, MATRIZEN& B, int mk);
    //
    // Unsymmetrische Koeffizientenmatrix nach Gauss staffeln
    //                                                 -> S3151(Kmat,Amat,Cmat,mk)
    // ===========================================================================
    // Unsymmetrische Koeffizientenmatrix
    // nach Gauss in die gestaffelte Dreiecksform ueberfuehren
    // K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // A(mk,mk) = Matrix fuer die gestaffelte Form der Obermatrix A
    //            (A darf gleich K sein) 
    // C(mk,mk) = Matrix fuer die gestaffelte Form der Untermatrix C
    //            (C darf gleich A sein)
    // mk       = Zeilen/Spalten von K, A, C
    // Ergebnis: Determinante der Matrix K
    // -------------------------------------------------------------
    double S3153(MATRIZEN& K, MATRIZEN& A, MATRIZEN& C, int mk);
    //
    // Rechte Seite nach Cholesky oder Gauss staffeln
    //                                     -> S3157(Rmat,Bmat,Ymat,mb,zr,symdef)
    // =========================================================================
    // Rechte Seite eines symmetrischen Gleichungssystems nach
    // Cholesky oder Gauss in die gestaffelte Rechteckform ueberfuehren
    // R(mb,zr) = Matrix mit voll abgespeicherten Rechten Seiten
    // B(mb,mb) = Matrix mit gestaffelter Koeffizientenmatrix
    //            (Untermatrix C bei Gauss unsymmetrisch)
    // Y(mb,zr) = Matrix fuer gestaffelte Form der Rechten Seiten
    //            (Y darf gleich R sein)
    // mb       = Zeilenzahl von B,R,Y
    // zr       = Spaltenzahl der Rechten Seiten R,Y
    // symdef= 1 : nach Cholesky staffeln
    //       = 0 : nach Gauss symmetrisch   staffeln
    //       =-1 : nach Gauss unsymmetrisch staffeln
    // ---------------------------------------------------------------------------
    void S3157(MATRIZEN& R, MATRIZEN& B, MATRIZEN& Y, int mb, int zr, int symdef = -1);
    //
    // Loesungen nach Cholesky oder Gauss ermitteln
    //                                       -> S3160(Xmat,Bmat,Ymat,mb,zr,symdef)
    // ===========================================================================
    // Loesungen fuer Gleichungsystem nach Cholesky (symmetrisch) oder
    // nach Gauss (symmetrisch oder unsymmetrisch) aus der gestaffelten
    // Koeffizientenmatrix und den gestaffelten Rechten Seiten ermitteln

    // X(mb,zr) = Matrix fuer voll abgespeicherte Loesungen
    // B(mb,mb) = Matrix mit gestaffelter Koeffizientenmatrix
    //            (Obermatrix A bei Gauss unsymmetrisch)
    // Y(mb,zr) = Matrix mit gestaffelter Form der Rechten Seiten
    //            (Y darf gleich X sein)
    // mb       = Zeilenzahl von B,X,Y
    // zr       = Spaltenzahl der Loesungen Seiten X,Y
    // symdef= 1 : nach Cholesky staffeln
    //       = 0 : nach Gauss symmetrisch   staffeln
    //       =-1 : nach Gauss unsymmetrisch staffeln
    // ----------------------------------------------------------------------------
    void S3160(MATRIZEN& X, MATRIZEN& B, MATRIZEN& Y, int mb, int zr, int symdef = -1);
    //
    // Inverse einer   symmetrischen Matrix nach Cholesky oder Gauss
    // Inverse einer unsymmetrischen Matrix nach Gauss               bilden
    //                                               -> S3170(Kmat,Bmat,mk,symdef)
    // ===========================================================================
    // Koeffizientenmatrix
    // nach Cholesky (nur positiv definite Matrix) oder Gauss invertieren
    // K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // B(mk,mk) = Matrix fuer die Inverse
    //            (B darf gleich K sein) 
    // mk       = Zeilen/Spalten von K,B
    // symdef   =  1 Cholesky
    //          =  0 Gauss (symmetrisch)
    //          = -1 Gauss (unsymmetrisch)
    // Ergebnis: Determinante der Matrix K
    // --------------------------------------------------------------------------
    double S3170(MATRIZEN& K, MATRIZEN& B, int mk, int symdef = -1);
    //
    // Lagerbedingungen in Koeffizientenmatrix einarbeiten
    //                                                -> S6115(Kmat,Lmat,Bmat,mk)
    // ==========================================================================
    // Lagerbedingungen in Koeffizientenmatrix einarbeiten 
    // K(mk,mk) = Matrix mit voll abgespeicherter Koeffizientenmatrix
    // L(mk)    = Vektor mit Lagerangaben
    //            -1 = Starre Lagerung
    //             0 = keine  Lagerung
    //            >0 = Federkonstante
    // B(mk,mk) = Matrix fuer die gelagerte Koeffizientenmatrix
    //            (B darf gleich K sein)
    // mk       = Zeilen/Spalten von K
    //            Zeilen von L
    // --------------------------------------------------------------------------
    void S6115(MATRIZEN& K_, MATRIZEN& L_, MATRIZEN& Bmat, int mk);
    //
    // ==========================================================================
    //                 Private Datenkomponenten -> Matrix-Deskriptor
    //                 =============================================
       private:
    // ========
    //
    // Anfangsindizes
    // --------------
    int ia,ja,ka,la;
    //
    // Typ des Einspeichern fuer den Zuweisungsoperator &operator = ()
    // --------------------
    //                ->  1 = Matrix
    //                ->  0 = Teil-Matrix  -> Funktion  teil_hinein(ia,ja,ka,la)
    //                -> -1 = Unter-Matrix -> Funktion unter_hinein(ia,ja,ka,la)
    int eintyp;
    };
    // ===========================================================================
    #endif