#ifndef rest_MATRIZEN_H
#define rest_MATRIZEN_H
// ================ Include-Files ==== ==============================
#include "pch.h"
// =============== Klassen-Deklaration: MATRIZEN  =======================================
// MATRIZEN-Basis-Klasse fuer allgemeine Matrizen vom Typ double
// Matrizen mat(i,j,k,l) werden spaltenweise gespeichert !!!
// Indizierung von i,j,k,l beginnt bei 0 !!!
// vgl. U. Breymann: C++. Carl Hanser Verlag (1993), S.186 ff
// ======================================================================================
// Copyright: Univ.-Prof. Dr.-Ing. Udo F. Meissner, Erkrath / Darmstadt
// experimentelles Lehr- und Studienprogramm - kommerzielle Nutzung nicht gestattet
// Der Autor bernimmt keine Haftung fr Unvollstndigkeiten und Maengel der Algorithmen!
// Version:  1995
// Compiler: C++
// ========= Deklaration der Prototypen (header file) =================
   class MATRIZEN
{
//    oeffentliche Mitgliedsfunktionen
//    ================================
//
// ******************************************************************** 
// *     ***** dieser Methodenteil ist im Buch dokumentiert *****     * 
// ********************************************************************
//
// ====================================================================
//    Geschtzte, virtuelle Mitgliedsfunktionen (Gleichungssysteme)
//    ===============================================================
   protected:
// ==========
//
// Symmetrische Koeffizientenmatrix Amat nach Gauss staffeln
// -> determinante = Amat.nach_gauss_staffeln(Umat)
//                   Umat = (Upper) Obermatrix 
// =====================================================
virtual double nach_gauss_staffeln(MATRIZEN& U) = 0;
//
// Symmetrische, positiv-definite Koeffizientenmatrix Amat nach Cholesky staffeln
// -> determinante = Amat.nach_cholesky_staffeln(Umat)
//                                               Umat = (Upper) Obermatrix 
// =========================================================================
virtual double nach_cholesky_staffeln(MATRIZEN& U) = 0;
//
// Unsymmetrische Koeffizientenmatrix Amat nach Gauss staffeln
// -> determinante = Amat.nach_gauss_staffeln(Umat,Lmat)
//                                            Umat = (Upper) Obermatrix 
//                                            Lmat = (Lower) Untermatrix 
// =====================================================================
virtual double nach_gauss_staffeln(MATRIZEN& U, MATRIZEN& L) = 0;
//
// Rechte Seite Rmat nach Cholesky oder Gauss staffeln 
// -> Rmat.staffeln(Lmat,Ymat,symdef)
//                  Lmat = (Lower) Untermatrix bei symdef = -1 Gauss (unsymmetrisch)
//                       =         Obermatrix bei  symdef =  0 Gauss (symmetrisch)
//                       =         Obermatrix bei  symdef =  1 Cholesky                 
//                       Ymat = gestaffelte Rechte Seiten
// =================================================================================
virtual void staffeln(MATRIZEN& L, MATRIZEN& Y, int symdef = -1) = 0;
//
// Loesungen Xmat nach Cholesky oder Gauss ermitteln
// -> Xmat.loesungen(Umat,Ymat,symdef)
//                   Umat = (Upper) Obermatrix bei symdef = -1 Gauss (unsymmetrisch)
//                        =         Obermatrix bei symdef =  0 Gauss (symmetrisch)
//                        =         Obermatrix bei symdef =  1 Cholesky                 
//                        Ymat = gestaffelte Rechte Seiten
// =================================================================================
virtual MATRIZEN& loesungen(MATRIZEN& U, MATRIZEN& Y, int symdef = -1) = 0;
//
// Inverse einer   symmetrischen Matrix Amat nach Cholesky oder Gauss bilden
// Inverse einer unsymmetrischen Matrix Amat nach Gauss bilden
// -> determinante = Amat.invertieren(Zmat,symdef)
//                                    Zmat = Inverse der Koeffizientenmatrix Amat
//                                         symdef  =  1 Cholesky
//                                                 =  0 Gauss (symmetrisch)
//                                                 = -1 Gauss (unsymmetrisch) 
// ==============================================================================
virtual double invertieren(MATRIZEN& Z, int symdef = -1) = 0;
//
// Lagerbedingungen in Koeffizientenmatrix Kmat einarbeiten
// -> Kmat.lagern(Svek, Amat)
//                Svek = Vektor mit Lagerangaben (Randbedingungen)
//                       Svekt(i) = -1 = Starre Lagerung
//                                   0 = keine  Lagerung
//                                >  0 = Federkonstante addiert zur Hauptdiagonalen
//                      Amat = gelagerte Koeffizientenmatrix (Typ MATRIX oder auch qMATRIX)
// ========================================================================================
virtual void lagern(MATRIZEN& S, MATRIZEN& A) = 0;
//
// ====================================================================
//                 Geschuetzte Mitgliedsfunktionen
//                 ===============================
   protected:
// ==========
//
// Skalarprodukt -> a*b -> zahl= a(s) * b(s)
// =========================================
double skalarprodukt(double *a,int da,double *b,int db,int n);
// ====================================================================
//                 Geschuetzte, rein virtuelle Mitgliedsfunktionen
//                 ===============================================
   protected:
// ==========
//
// Adresse von Matrixelement -> mat.adresse(i,j,k,l)
//==================================================
virtual double *adresse(int i=0,int j=0,int k=0,int l=0)=0;
//
// Matrix erzeugen -> erzeugen(ni,nj,nk,nl,neutyp,C_array)
// =======================================================
// bei neutyp=-1 temporaere Matrix erzeugen
// bei neutyp= 1 permanente Matrix erzeugen
// C_array vorinitialisiertes C_Feld [] = { ... }
// ----------------------------------------
virtual void erzeugen(int ni=1,int nj=1,int nk=1,int nl=1,int neutyp=1,
                                                   double* C_array = NULL)=0;
//
// Matrix kopieren -> kopieren(&mat,neutyp)
// ========================================
// bei neutyp=-1 Matrix-Deskriptor, neuen Speicher anfordern (nur privat !!!)
// bei neutyp= 0 Matrix-Deskriptor, alten Speicher benutzen  (nur privat !!!)
// bei neutyp= 1 Matrix-Deskriptor, neuen Speicher anfordern, Matrix kopieren
// ---------------------------------------------------------------------------
virtual void kopieren(MATRIZEN &mat,int neutyp=1)=0;
//
// temporaere Matrix erzeugen -> temporaere_matrix_erzeugen(ni,nj,nk,nl,neutyp)
// ============================================================================
// ni, nj, nk, nl: Dimensionen der 4D-Matrix
// bei neutyp=-1 temporaere Matrix erzeugen
// bei neutyp= 1 permanente Matrix erzeugen
// -----------------------------------------
virtual MATRIZEN *temporaere_matrix_erzeugen(int ni,int nj=1,int nk=1,
                                                     int nl=1,int neutyp=-1)=0;
//
// temporaere Matrix als Kopie erzeugen
//                                  -> temporaere_matrix_erzeugen(&mat,neutyp)
// ===========================================================================
// Kopie der Matrix mat:
// bei neutyp=-1 Matrix-Deskriptor, neuen Speicher anfordern (nur privat !!!)
// bei neutyp= 0 Matrix-Deskriptor, alten Speicher benutzen  (nur privat !!!)
// bei neutyp= 1 Matrix-Deskriptor, neuen Speicher anfordern, Matrix kopieren
// ---------------------------------------------------------------------------
virtual MATRIZEN *temporaere_matrix_erzeugen(MATRIZEN &mat,int neutyp=-1)=0;
//
// Matrix-Dimensionen temporaer aendern -> mat.dimension_temporaer(i,j,k,l)
// ===========================================================================
// Das Zusammenfassen von Indizes (Umdimensionierung) wird eigentlich nur
//                    fuer die Matrizenmultiplikation mat1*mat2 benoetigt.
// Daher gilt die Umdimensionierung nur bis zur naechsten Matrizenoperartion !
// Danach werden die Originaldimensionen wiederhergestellt !!!
// ---------------------------------------------------------------------------
virtual void dimension_temporaer(int ni=1,int nj=1,int nk=1,int nl=1)=0;
//
// Teilmatrix heraus -> ...=mat1.teil_heraus(mat2, i,j,k,l)
// =========================================================
// teil_heraus holt ab Index mat2(i,j,k,l) aufeinanderfolgende Matrixelemente
// heraus, die in der Teilmatrix mat1(ni,nj,nk,nl) gespeichert werden
// ---------------------------------------------------------------------------
virtual void teil_heraus(MATRIZEN &mat2, int i=0,int j=0,int k=0,int l=0)=0;
//
// Teilmatrix hinein -> mat.teil_hinein(i,j,k,l)=...
// =================================================
// teil_hinein erlaubt das Einspeichern der Teilmatrix ab Index mat(i,j,k,l)
// ---------------------------------------------------------------------------
virtual void teil_hinein(int i=0, int j=0, int k=0, int l=0)=0;
//
// Untermatrix heraus -> ...=mat1.unter_heraus(mat2, i,j,k,l)
// ==========================================================
// unter_heraus holt ab Index mat2(i,j,k,l)
//                                    die Untermatrix mat1(ni,nj,nk,nl) heraus
// ---------------------------------------------------------------------------
virtual void unter_heraus(MATRIZEN &mat1, int i=0,int j=0,int k=0,int l=0)=0;
//
// Untermatrix hinein -> mat.unter_hinein(i,j,k,l)=...
// ===================================================
// unter_hinein erlaubt das Einspeichern der Untermatrix ab Index mat(i,j,k,l)
// ---------------------------------------------------------------------------
virtual void unter_hinein(int i=0,int j=0,int k=0,int l=0)=0;
//
// Matrix-Ausgabe -> mat.ausgeben(s)
// =================================
virtual void ausgeben(std::ostream &s)=0;
//
// Matrix-Eingabe -> mat.einlesen(s)
// =================================
virtual void einlesen(std::istream &s)=0;
//
// Alle Matrix-Elemente belegen -> mat.zuweisen(zahl,zweig)
//                                 (nur Matrizeninhalt kopieren)
// =============================================================
// Zweig= 0 -> mat=    zahl
//      = 2 -> mat=mat+zahl
//      = 4 -> mat=mat-zahl
// ------------------------
virtual void zuweisen(double zahl,int zweig=0)=0;
//
// Matrix zuweisen -> mat1.zuweisen(mat2,zweig)
//                    (nur Matrizeninhalt kopieren)
// ================================================
// Zweig= 0 -> mat1=    mat2
//      = 2 -> mat1=mat+mat2
//      = 4 -> mat1=mat-mat2
virtual void zuweisen(MATRIZEN &mat2,int zweig=0)=0;
//
// Addition -> mat1+mat2 bzw. mat1+vek2 bzw. vek1+mat2
// ===================================================
virtual void addieren(MATRIZEN &mat1,MATRIZEN &mat2)=0;
//
// Subtraktion -> mat1-mat2 bzw. mat1-vek2 bzw. vek1-mat2
// ======================================================
virtual void subtrahieren(MATRIZEN &mat1,MATRIZEN &mat2)=0;
//
// Matrizenmultiplikation -> mat1*mat2
// -> c(i,j,k,l)=mat1(i,s,j) * mat2(k,s,l)
// =====================================================================
virtual void multiplizieren(MATRIZEN &mat1,MATRIZEN &mat2,int zweig=0)=0;
//
// Matrizendivision -> matx(s,n) = matr(i,n) / MATk(i,s)
//         MATk(i,s) * matx(s,n) = matr(i,n)
// =====================================================
virtual MATRIZEN& dividieren(MATRIZEN &matr) = 0;
//
// =====================================================================
//                 Geschuetzte Datenkomponenten -> Matrix-Deskriptor
//                 =================================================
// 
// *********************************************************************
// *      ***** dieser Datenteil ist im Buch dokumentiert *****        *
// *********************************************************************
}; 
// =====================================================================
#endif